/// @file MultiUnion.h
/// @author mihaela.gheata@cern.ch
//
/// Includes all headers related to the multi-union volume.

#ifndef VECGEOM_VOLUMES_MULTIUNION_H_
#define VECGEOM_VOLUMES_MULTIUNION_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedMultiUnion.h"
#include "VecGeom/volumes/SpecializedMultiUnion.h"
#include "VecGeom/volumes/UnplacedMultiUnion.h"

#endif // VECGEOM_VOLUMES_MULTIUNION_H_
