/// @file Hype.h

#ifndef VECGEOM_VOLUMES_HYPE_H_
#define VECGEOM_VOLUMES_HYPE_H_

#include "VecGeom/base/Global.h"

#include "VecGeom/volumes/PlacedHype.h"
#include "VecGeom/volumes/SpecializedHype.h"
#include "VecGeom/volumes/UnplacedHype.h"

#endif // VECGEOM_VOLUMES_HYPE_H_
