/// \file Polyhedron.h
/// \author Johannes de Fine Licht (johannes.definelicht@cern.ch)

#ifndef VECGEOM_VOLUMES_POLYHEDRON_H_
#define VECGEOM_VOLUMES_POLYHEDRON_H_

#include "VecGeom/volumes/UnplacedPolyhedron.h"
#include "VecGeom/volumes/PlacedPolyhedron.h"
#include "VecGeom/volumes/SpecializedPolyhedron.h"

#endif // VECGEOM_VOLUMES_POLYHEDRON_H_
