/// @file: Trapezoid.h
/// @author Guilherme Lima (lima 'at' fnal 'dot' gov)
//
//   Include all headers related to the trapezoid shape
//
//  2014-05-01 Guilherme Lima  Created, based on the box example
//  2016-07-25 Guilherme Lima  Converted to new implementation helpers + VecCore
//

#ifndef VECGEOM_VOLUMES_TRAPEZOID_H_
#define VECGEOM_VOLUMES_TRAPEZOID_H_

#include "VecGeom/base/Global.h"

#include "VecGeom/volumes/PlacedTrapezoid.h"
#include "VecGeom/volumes/SpecializedTrapezoid.h"
#include "VecGeom/volumes/UnplacedTrapezoid.h"

#endif // VECGEOM_VOLUMES_TRAPEZOID_H_
