delete from Arb8;
delete from BooleanSolid;
delete from Box;
delete from Cone;
delete from Daughter_Volume;
delete from Element;
delete from ElementsInMaterial;
delete from Isotope;
delete from IsotopesInElement;
delete from Material;
delete from Polycone;
delete from Polycone_plane;
delete from Polyhedra;
delete from Polyhedra_plane;
delete from StressTest;
delete from TestedVolume;
delete from Torus;
delete from Transformation;
delete from Trap;
delete from Trd;
delete from Tube;
delete from Volume;
alter table Material auto_increment=1;
alter table Element auto_increment=1;
alter table Isotope auto_increment=1;