
/// @file Cone.h
//
/// Includes all headers related to the cone volume

#ifndef VECGEOM_VOLUMES_CONE_H_
#define VECGEOM_VOLUMES_CONE_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedCone.h"
#include "VecGeom/volumes/SpecializedCone.h"
#include "VecGeom/volumes/UnplacedCone.h"

#endif // VECGEOM_VOLUMES_CONE_H_
