/*
 * Polycone.h
 *
 *  Created on: Dec 8, 2014
 *      Author: swenzel
 */

#ifndef VECGEOM_VOLUMES_POLYCONE_H_
#define VECGEOM_VOLUMES_POLYCONE_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedPolycone.h"
#include "VecGeom/volumes/SpecializedPolycone.h"
#include "VecGeom/volumes/UnplacedPolycone.h"

#endif /* VECGEOM_VOLUMES_POLYCONE_H_ */
