/// @file Box.h
/// @author Johannes de Fine Licht (johannes.definelicht@cern.ch)
//
/// Includes all headers related to the box volume.

#ifndef VECGEOM_VOLUMES_BOX_H_
#define VECGEOM_VOLUMES_BOX_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedBox.h"
#include "VecGeom/volumes/SpecializedBox.h"
#include "VecGeom/volumes/UnplacedBox.h"

#endif // VECGEOM_VOLUMES_BOX_H_
