
/// @file ScaledShape.h
/// @author Mihaela Gheata (mihaela.gheata@cern.ch)
//
/// Includes all headers related to the tube volume

#ifndef VECGEOM_VOLUMES_SCALEDSHAPE_H_
#define VECGEOM_VOLUMES_SCALEDSHAPE_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedScaledShape.h"
#include "VecGeom/volumes/SpecializedScaledShape.h"
#include "VecGeom/volumes/UnplacedScaledShape.h"

#endif // VECGEOM_VOLUMES_SCALEDSHAPE_H_
