/// @file Torus2.h
/// Includes all headers related to the torus volume.

#ifndef VECGEOM_VOLUMES_TORUS2_H_
#define VECGEOM_VOLUMES_TORUS2_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedTorus2.h"
#include "VecGeom/volumes/SpecializedTorus2.h"
#include "VecGeom/volumes/UnplacedTorus2.h"

#endif // VECGEOM_VOLUMES_TORUS2_H_
