/// @file GenericPolycone.h
/// @author Raman Sehgal (raman.sehgal@cern.ch)

#ifndef VECGEOM_VOLUMES_GENERICPOLYCONE_H_
#define VECGEOM_VOLUMES_GENERICPOLYCONE_H_

#include "VecGeom/base/Global.h"

#include "VecGeom/volumes/PlacedGenericPolycone.h"
#include "VecGeom/volumes/SpecializedGenericPolycone.h"
#include "VecGeom/volumes/UnplacedGenericPolycone.h"

#endif // VECGEOM_VOLUMES_GENERICPOLYCONE_H_
