/// \file NavigationState.h
/// \author Sandro Wenzel (sandro.wenzel@cern.ch)
/// \date 12.03.2014

#ifndef VECGEOM_NAVIGATION_NAVIGATIONSTATE_H_
#define VECGEOM_NAVIGATION_NAVIGATIONSTATE_H_

#include "VecGeom/navigation/NavStateFwd.h"
#include "VecGeom/navigation/NavStatePath.h"
#include "VecGeom/navigation/NavStateIndex.h"

#endif // VECGEOM_NAVIGATION_NAVIGATIONSTATE_H_
