
/// @file Tube.h
/// @author Georgios Bitzes (georgios.bitzes@cern.ch)
//
/// Includes all headers related to the tube volume

#ifndef VECGEOM_VOLUMES_TUBE_H_
#define VECGEOM_VOLUMES_TUBE_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedTube.h"
#include "VecGeom/volumes/SpecializedTube.h"
#include "VecGeom/volumes/UnplacedTube.h"

#endif // VECGEOM_VOLUMES_TUBE_H_
