// This file is part of VecGeom and is distributed under the
// conditions in the file LICENSE.txt in the top directory.
// For the full list of authors see CONTRIBUTORS.txt and `git log`.
/// @file volumes/Tessellated.h
/// @author Mihaela Gheata
///
/// Includes all headers related to the tessellated volume.

#ifndef VECGEOM_VOLUMES_TESSELLATED_H_
#define VECGEOM_VOLUMES_TESSELLATED_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedTessellated.h"
#include "VecGeom/volumes/SpecializedTessellated.h"
#include "VecGeom/volumes/UnplacedTessellated.h"

#endif // VECGEOM_VOLUMES_TESSELLATED_H_
