/*
 * CutTube.h
 *
 *  Created on: 03.11.2016
 *      Author: mgheata
 */
/// Includes all headers related to the cut tube volume

#ifndef VECGEOM_VOLUMES_CUTTUBE_H_
#define VECGEOM_VOLUMES_CUTTUBE_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedCutTube.h"
#include "VecGeom/volumes/SpecializedCutTube.h"
#include "VecGeom/volumes/UnplacedCutTube.h"

#endif // VECGEOM_VOLUMES_CUTTUBE_H_
