/// Includes all headers related to the simple extruded volume.

#ifndef VECGEOM_VOLUMES_SEXTRU_H_
#define VECGEOM_VOLUMES_SEXTRU_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedSExtru.h"
#include "VecGeom/volumes/SpecializedSExtru.h"
#include "VecGeom/volumes/UnplacedSExtruVolume.h"

#endif // VECGEOM_VOLUMES_SEXTRU_H_
