/// @file Extruded.h
/// @author mihaela.gheata@cern.ch
//
/// Includes all headers related to the extruded volume.

#ifndef VECGEOM_VOLUMES_EXTRUDED_H_
#define VECGEOM_VOLUMES_EXTRUDED_H_

#include "VecGeom/base/Global.h"
#include "VecGeom/volumes/PlacedExtruded.h"
#include "VecGeom/volumes/SpecializedExtruded.h"
#include "VecGeom/volumes/UnplacedExtruded.h"

#endif // VECGEOM_VOLUMES_EXTRUDED_H_
